/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is a test framework for NEON YUV-to-ARGB code.
 *
 * The Initial Developer of the Original Code is ARM Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jacob Bramley <Jacob.Bramley@arm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "framework.h"

#include "bmp_writer.h"

int write_bmp(char const * name, void const * data, uint32_t width, uint32_t height, uint32_t bpp)
{
    int  ret = 0;

    uint8_t const * d = data;
    uint32_t bmp_row_size;

    struct {
        char     magic[2];
        uint32_t size;
        uint16_t reserved0;
        uint16_t reserved1;
        uint32_t offset;
    } bmp_header;
    uint32_t const bmp_header_size = 14;    //2+4+2+2+4;

    struct {
        uint32_t size;          // Size of the header, not the file!
        uint32_t width;
        uint32_t height;
        uint16_t planes;        // Must be 1.
        uint16_t bpp;           // Use 16 for RGB-565.
        uint32_t compression;   // Set to 0 for no compression.
        uint32_t isize;         // Raw bitmap data size. Can be 0 for no compression.
        int32_t  xres;          // Pixels per meter.
        int32_t  yres;          // Pixels per meter.
        uint32_t colours;       // Number of colours in the palette. (Set to 0.)
        uint32_t icolours;      // Number of important colours in the palette. (Set to 0.)
    } bmp_dib;
    uint32_t const bmp_dib_size = 40;       // 4+4+4+2+2+4+4+4+4+4+4

    FILE * f = fopen(name, "wb");
    if (!f) {
        printf("Error: Could not create BMP file \"%s\".\n", name);
        goto exit;
    }

    // Work out the BMP row size (including padding to 4-byte boundaries).
    bmp_row_size = (width * 3 + 0x3) & ~0x3;

    memcpy(bmp_header.magic, "BM", 2);
    bmp_header.size = bmp_header_size + bmp_dib_size + bmp_row_size * height;
    bmp_header.reserved0 = 0;
    bmp_header.reserved1 = 0;
    bmp_header.offset = bmp_header_size + bmp_dib_size;

    bmp_dib.size = bmp_dib_size;
    bmp_dib.width = width;
    bmp_dib.height = height;
    bmp_dib.planes = 1;
    bmp_dib.bpp = 24;
    bmp_dib.compression = 0;
    bmp_dib.isize = bmp_row_size * height;
    bmp_dib.xres = 2835;        // Fairly arbitrary.
    bmp_dib.yres = 2835;        // Fairly arbitrary.
    bmp_dib.colours = 0;
    bmp_dib.icolours = 0;

    // Write bitmap file header.
    fwrite(&bmp_header.magic, 1, 2, f);
    fwrite(&bmp_header.size, 4, 1, f);
    fwrite(&bmp_header.reserved0, 2, 1, f);
    fwrite(&bmp_header.reserved1, 2, 1, f);
    fwrite(&bmp_header.offset, 4, 1, f);

    assert(bmp_header_size == ftell(f));

    // Write a simple 40-byte DIB "BITMAPINFOHEADER" section.
    fwrite(&bmp_dib.size, 4, 1, f);
    fwrite(&bmp_dib.width, 4, 1, f);
    fwrite(&bmp_dib.height, 4, 1, f);
    fwrite(&bmp_dib.planes, 2, 1, f);
    fwrite(&bmp_dib.bpp, 2, 1, f);
    fwrite(&bmp_dib.compression, 4, 1, f);
    fwrite(&bmp_dib.isize, 4, 1, f);
    fwrite(&bmp_dib.xres, 4, 1, f);
    fwrite(&bmp_dib.yres, 4, 1, f);
    fwrite(&bmp_dib.colours, 4, 1, f);
    fwrite(&bmp_dib.icolours, 4, 1, f);

    assert(bmp_dib_size + bmp_header_size == ftell(f));
    assert(bmp_header.offset == ftell(f));

    // Get bytes per pixel.
    assert((bpp == 16) || (bpp == 24) || (bpp == 32));
    bpp /= 8;

    for(uint32_t y = height; y > 0; --y) {
        // Seek to the start of the row. Note that bitmaps are stored
        // upside-down, so we write to the end of the file first.
        fseek(f, bmp_header.offset + bmp_row_size * (y-1), SEEK_SET);

        switch (bpp) {
          case 3:   // RGB-888
            // This format requires no conversion, so write whole rows.
            fwrite(d, 3, width, f);
            d += width*3;
            break;
          case 2:   // RGB-565
            // This format requires up-conversion.
            for (uint32_t x = 0; x < width; x++, d+=2) {
                uint32_t    rgb;
                uint32_t    r;
                uint32_t    g;
                uint32_t    b;
                memcpy(&rgb, d, 2);
                // Extract the 5- and 6-bit fields.
                r = (rgb >> 11) & 0x1f; // 5 bits.
                g = (rgb >>  5) & 0x3f; // 6 bits.
                b = (rgb >>  0) & 0x1f; // 5 bits.
                // Extend to 8-bit fields, and duplicate the top bit in the
                // bottom bits to keep the whites white.
                r = (r << 3) | (r >> 2);
                g = (g << 2) | (g >> 4);
                b = (b << 3) | (b >> 2);
                // Write the data to the BMP file.
                rgb = (r << 16) | (g << 8) | (b << 0);
                fwrite(&rgb, 3, 1, f);
            }
            break;
          case 4:   // ARGB-8888
            // This format requires trivial down-conversion by dropping A.
            for (uint32_t x = 0; x < width; x++, d+=4) {
                uint32_t    argb;
                uint32_t    r;
                uint32_t    g;
                uint32_t    b;
                memcpy(&argb, d, 4);
                // Extract the relevant fields.
                r = (argb >> 16) & 0xff;
                g = (argb >>  8) & 0xff;
                b = (argb >>  0) & 0xff;
                // Write the data to the BMP file.
                argb = (r << 16) | (g << 8) | (b << 0);
                fwrite(&argb, 3, 1, f);
            }
            break;
        }

        {
            uint8_t zero[] = {0,0,0,0};
            fwrite(zero, 1, bmp_row_size - width*3, f);
        }
    }

    fseek(f, 0, SEEK_END);
    assert(bmp_header.size == ftell(f));

    fflush(f);

    ret = 1;
exit:
    if (f) {
        fclose(f);
    }
    return ret;
}

