/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is a test framework for NEON YUV-to-ARGB code.
 *
 * The Initial Developer of the Original Code is ARM Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jacob Bramley <Jacob.Bramley@arm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef FRAMEWORK_H
#define FRAMEWORK_H

#include <stdint.h>

typedef enum {
    FMT_IYUV,   FMT_I420=FMT_IYUV,  // Planar, 2x2 subsampled V and U.
    FMT_YV12,                       // Planar, 2x2 subsampled U and V.
    FMT_YV16,                       // Planar, 2x1 subsampled U and V.
    FMT_YV24                        // Planar, 1x1 subsampled U and V.
} format_t;

typedef struct {
    char *          base_name;  // This is allocated with malloc and must be freed.
    uint32_t        base_name_len;
    char const *    src_file;
    format_t        format;
    uint32_t        src_width;
    uint32_t        src_height;
    uint32_t        pad_top;
    uint32_t        pad_right;
    uint32_t        pad_bottom;
    uint32_t        pad_left;
    uint32_t        dst_width;  // Output canvas size, not scaled image size!
    uint32_t        dst_height; // Output canvas size, not scaled image size!
    uint32_t        dst_x;
    uint32_t        dst_y;
    float           scale_x;
    float           scale_y;
    uint32_t        runs;
    int             verbose;
} options_t;

// -------- Utilities --------

static inline void * alignUp(void * in, uint32_t align) {
    uintptr_t p = (uintptr_t)in;
    return (void*)((p + (align-1)) & ~(align-1));
}

#define OOM_CHECK(ptr)  do { if (!ptr) { printf("Out of memory at line %u.\n", __LINE__); goto exit; } } while(0);

// -------- Test Function Prototypes --------

typedef void (*scale_rgb565_t) (
        uint8_t const * y,
        uint8_t const * u,
        uint8_t const * v,
        void          * rgb,
        int             src_left,
        int             src_top,
        int             src_width,
        int             src_height,
        int             dst_width,
        int             dst_height,
        int             y_pitch,    // Source row-stride.
        int             uv_pitch,   // Source row-stride for U and V.
        int             rgb_pitch,  // Destination row-stride.
        format_t        src_format);

#define DECLARE_SCALE_YCBCR_RGB565_BILINEAR(func)   \
extern void func(                                                       \
        uint8_t const * y,                                              \
        uint8_t const * u,                                              \
        uint8_t const * v,                                              \
        void          * rgb,                                            \
        int             src_left,                                       \
        int             src_top,                                        \
        int             src_width,                                      \
        int             src_height,                                     \
        int             dst_width,                                      \
        int             dst_height,                                     \
        int             y_pitch,    /* Source row-stride.            */ \
        int             uv_pitch,   /* Source row-stride for U and V.*/ \
        int             rgb_pitch,  /* Destination row-stride.       */ \
        format_t        src_format)

DECLARE_SCALE_YCBCR_RGB565_BILINEAR(    ScaleYCbCrToRGB565_Bilinear_C                           );
DECLARE_SCALE_YCBCR_RGB565_BILINEAR(    ScaleYCbCrToRGB565_Bilinear_NULL                        );
DECLARE_SCALE_YCBCR_RGB565_BILINEAR(    ScaleYCbCrToRGB565_Bilinear_NEON_jbramley               );
DECLARE_SCALE_YCBCR_RGB565_BILINEAR(    ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_loadloop      );
DECLARE_SCALE_YCBCR_RGB565_BILINEAR(    ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_vtbl          );
DECLARE_SCALE_YCBCR_RGB565_BILINEAR(    ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_vtbl_opt      );
DECLARE_SCALE_YCBCR_RGB565_BILINEAR(    ScaleYCbCrToRGB565_Bilinear_NEON_derf                   );

#endif  // FRAMEWORK_H

