// Convert one or more BMP files (in the YUV colour space) to RAW YUV (YV12) format.

#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

char const * me;    // The name of this program (taken from argv[0]).

// ---- Conversion routines. ----

// Note that the values in this enum are used to loop up names in the
// format_names array.
typedef enum {
    FMT_IYUV=0, FMT_I420=FMT_IYUV,      // I420 and IYUV are identical.
    FMT_YV12=1,
} out_format_t;

char const * const format_names[] = {
    "iyuv",     // FMT_IYUV, FMT_I420
    "yv12"      // FMT_YV12
};

// Pad with bright pink, so it stands out in case a YUV reader reads it.
uint8_t pad_y  = 128;
uint8_t pad_u  = 255;
uint8_t pad_v  = 255;

void rgb_to_yuv(char const * in_name, out_format_t out_format, uint32_t padding)
{
    FILE * in;
    FILE * out;
    uint32_t const base_name_length = strrchr(in_name, '.') - in_name;
    // The output name is: "<base_name>-<width>x<height>[-pad<pad>].<format>"
    char * out_name = malloc(base_name_length + 1 + 10 + 1 + 10 + 1 + 3 + 10 + 1 + 4 + 1);
    void * buffer = NULL;

    uint32_t width = 0;
    uint32_t height = 0;
    uint32_t dst_width = 0;
    uint32_t dst_height = 0;
    uint32_t bpp = 0;
    uint32_t data_offset = 0;
    uint32_t dib_size = 0;
    uint32_t tmp;
    uint32_t bmp_row_size;

    uint8_t *   out_y = NULL;
    uint8_t *   out_u = NULL;
    uint8_t *   out_v = NULL;
    uint32_t    out_offset = 0;

    if (!out_name) {
        printf("Error: Out of memory at line %u.\n", __LINE__);
        goto exit;
    }

    if (padding % 2) {
        printf("Error: Padding size must be a multiple of 2 for this format.\n");
        goto exit;
    }

    in = fopen(in_name, "rb");
    if (!in) {
        printf("Error: Could not open input file \"%s\".\n", in_name);
        goto exit;
    }
    
    // Get the data offset from the bitmap file header.
    fseek(in, 10, SEEK_SET);
    fread(&data_offset, 4, 1, in);

    // Get the image size and colour depth from the DIB header.
    fseek(in, 14, SEEK_SET);
    fread(&dib_size, 4, 1, in);
    switch (dib_size) {
      case 40:  // BITMAPINFOHEADER
        fread(&width, 4, 1, in);
        fread(&height, 4, 1, in);
        fseek(in, 28, SEEK_SET);
        fread(&bpp, 2, 1, in);
        // Check there's no compression in use.
        fread(&tmp, 4, 1, in);
        if (tmp != 0) {
            printf("Error: Unsupported compression type (%u).\n", tmp);
            goto exit;
        }
        // Check there's no colour palette.
        fseek(in, 46, SEEK_SET);
        fread(&tmp, 4, 1, in);
        if (tmp != 0) {
            printf("Error: No colour palette support.\n");
        }
        break;
      case 12:  // BITMAPCOREHEADER
      case 64:  // BITMAPCOREHEADER2
      default:
        printf("Error: Unsupported bitmap DIB header type. (Size is %u bytes.)\n", dib_size);
        goto exit;
    }

    if ((out_format == FMT_YV12) && ((width % 2) || (height % 2))) {
        printf("Error: Width (%u) and height (%u) must both be even for YV12.\n", width, height);
        goto exit;
    }

    // Work out how big the output will be.
    dst_width = width + 2*padding;
    dst_height = height + 2*padding;

    // Check that the colour depth is supported.
    switch (bpp) {
      case 16:
      case 24:
      case 32:
        break;
      default:
        printf("Error: Unsupported bitmap colour depth (%u bpp).\n", bpp);
        goto exit;
    }

    // Convert bits per pixel to bytes per pixel.
    bpp /= 8;

    // Allocate an output buffer.
    out_y = malloc(dst_width*dst_height);
    out_u = malloc(dst_width*dst_height);
    out_v = malloc(dst_width*dst_height);

    if ((!out_y) || (!out_u) || (!out_v)) {
        printf("Error: Out of memory at line %u.\n", __LINE__);
        goto exit;
    }

    // Work out how big the stored BMP rows are (including their padding).
    bmp_row_size = (width * bpp + 0x3) & ~0x3;

    // Print top padding.
    for (uint32_t y = 0; y < padding; y++) {
        for (uint32_t x = 0; x < dst_width; ++x) {
            out_y[out_offset] = pad_y;
            out_u[out_offset] = pad_u;
            out_v[out_offset] = pad_v;
            out_offset++;
        }
    }

    for (uint32_t y = 0; y < height; ++y) {
        // Flip the image on loading. BMP files are stored upside-down.
        fseek(in, data_offset + bmp_row_size * (height - y - 1), SEEK_SET);

        // Print left padding.
        for (uint32_t x = 0; x < padding; ++x) {
            out_y[out_offset] = pad_y;
            out_u[out_offset] = pad_u;
            out_v[out_offset] = pad_v;
            out_offset++;
        }

        for (uint32_t x = 0; x < width; ++x) {
            uint32_t    r;
            uint32_t    g;
            uint32_t    b;

            if (!fread(&tmp, bpp, 1, in)) {
                printf("Error: Could not read pixel [%u,%u] from \"%s\".\n", x, y, in_name);
                goto exit;
            }
            switch (bpp) {
              case 2:   // RGB-555
                // Extract the 5-bit fields.
                r = (tmp >> 10) & 0x1f;
                g = (tmp >>  5) & 0x1f;
                b = (tmp >>  0) & 0x1f;
                // Extend to 8-bit fields, and duplicate the top bit in the
                // bottom bits to keep the whites white.
                r = (r << 3) | (r >> 2);
                g = (g << 3) | (g >> 2);
                b = (b << 3) | (b >> 2);
                break;
              case 3:   // RGB-888
              case 4:   // ARGB-8888
                r = (tmp >> 16) & 0xff;
                g = (tmp >>  8) & 0xff;
                b = (tmp >>  0) & 0xff;
                // Ignore alpha (for ARGB).
                break;
              default:
                assert(0);
            }

            // Convert the pixel to YCbCr.
            float y_f =  16.0f + ( 65.738f*r + 129.057f*g +  25.064f*b) / 256.0f;
            float u_f = 128.0f + (-37.956f*r + -74.494f*g + 112.439f*b) / 256.0f;
            float v_f = 128.0f + (112.439f*r + -94.154f*g + -18.285f*b) / 256.0f;

            out_y[out_offset] = (uint8_t)(fminf(255.0f, fmaxf(0.0f, y_f)));
            out_u[out_offset] = (uint8_t)(fminf(255.0f, fmaxf(0.0f, u_f)));
            out_v[out_offset] = (uint8_t)(fminf(255.0f, fmaxf(0.0f, v_f)));
            out_offset++;
        }

        // Print right padding.
        for (uint32_t x = 0; x < padding; ++x) {
            out_y[out_offset] = pad_y;
            out_u[out_offset] = pad_u;
            out_v[out_offset] = pad_v;
            out_offset++;
        }
    }
    fclose(in);

    // Print bottom padding.
    for (uint32_t y = 0; y < padding; y++) {
        for (uint32_t x = 0; x < dst_width; ++x) {
            out_y[out_offset] = pad_y;
            out_u[out_offset] = pad_u;
            out_v[out_offset] = pad_v;
            out_offset++;
        }
    }

    // Work out what the output filename will be.
    memcpy(out_name, in_name, base_name_length);
    if (padding) {
        sprintf(out_name+base_name_length, "-%ux%u-pad%u.%s", width, height, padding, format_names[out_format]);
    } else {
        sprintf(out_name+base_name_length, "-%ux%u.%s", width, height, format_names[out_format]);
    }

    out = fopen(out_name, "wb");
    if (!out) {
        printf("Error: Could not create output file \"%s\".\n", out_name);
        goto exit;
    }

    switch(out_format) {
      case FMT_YV12:
      case FMT_IYUV:
        // This is a planar format. Y' is encoded first, followed by U and V.
        // The V and U channels cover 2x2 blocks. In YV12, V comes first. In
        // IYUV (and I420), U comes first.
        fwrite(out_y, 1, dst_width*dst_height, out);
        for (uint32_t y = 0; y < dst_height/2; ++y) {
            for (uint32_t x = 0; x < dst_width/2; ++x) {
                uint32_t    new_u = 0;
                new_u += out_u[(y*2  )*dst_width + x*2  ];
                new_u += out_u[(y*2  )*dst_width + x*2+1];
                new_u += out_u[(y*2+1)*dst_width + x*2  ];
                new_u += out_u[(y*2+1)*dst_width + x*2+1];
                out_u[y*dst_width/2+x] = new_u / 4;

                uint32_t    new_v = 0;
                new_v += out_v[(y*2  )*dst_width + x*2  ];
                new_v += out_v[(y*2  )*dst_width + x*2+1];
                new_v += out_v[(y*2+1)*dst_width + x*2  ];
                new_v += out_v[(y*2+1)*dst_width + x*2+1];
                out_v[y*dst_width/2+x] = new_v / 4;
            }
        }
        if (out_format == FMT_YV12) {
            fwrite(out_v, 1, dst_width*dst_height/4, out);
            fwrite(out_u, 1, dst_width*dst_height/4, out);
        } else {    // out_format == FMT_IYUV
            fwrite(out_u, 1, dst_width*dst_height/4, out);
            fwrite(out_v, 1, dst_width*dst_height/4, out);
        }
        break;
      default:
        assert(0);
    }

    fclose(out);
    printf("Wrote file: %s\n", out_name);

exit:
    free(out_y);
    free(out_u);
    free(out_v);
    free(out_name);
    free(buffer);
}



void usage()
{
    printf("Usage: %s [options] [source file list ...]\n", me);
    printf("Options:\n");
    printf(" --out-format=<format>\n");
    printf(" -o <format>\n");
    printf("     Specify the output format. The default is IYUV. Supported formats:\n");
    printf("         IYUV\n");
    printf("         I420   Note: I420 is equivalent to IYUV.\n");
    printf("         YV12\n");
    printf("     Refer to http://www.fourcc.org/ for a description of the formats.\n");
    printf(" --padding=<padding>\n");
    printf(" -p <padding>\n");
    printf("     Specify the width of a border which will be inserted around the edge of\n");
    printf("     the image. This can be used to test row-stride functions in YUV readers.\n");
    printf(" --help\n");
    printf(" -h\n");
    printf("     Show this usage information.\n");
}

static int parse_out_format(char const * str, out_format_t * fmt)
{
    if (strcmp(str, "YV12") == 0) {
        *fmt = FMT_YV12;
        return 1;
    }
    if ((strcmp(str, "IYUV") == 0) || (strcmp(str, "I420") == 0)) {
        *fmt = FMT_IYUV;
        return 1;
    }
    printf("Error: Unrecognized output format \"%s\".\n", str);
    usage();
    return 0;
}

int confirm_remaining_arguments(char const * option, uint32_t count, uint32_t arg, int argc)
{
    arg += count;
    if (arg >= argc) {
        printf("Error: The \"%s\" option requires an argument.\n", option);
        usage();
        return 0;
    }
    return 1;
}

int main(int argc, char const * argv[])
{
    int             ret = EXIT_FAILURE;
    char const **   sources = NULL;
    uint32_t        source_count = 0;

    out_format_t    out_format = FMT_IYUV;
    uint32_t        padding = 0;

    me = argv[0];

    // Allocate a source file name pointer for every argument, even though some
    // will be unused.
    sources = malloc(sizeof(argv[0]) * (argc-1));
    if (!sources) {
        printf("Error: Out of memory at line %u.\n", __LINE__);
        goto exit;
    }

    // Parse arguments.
    for (uint32_t arg = 1; arg < argc; ++arg) {
        
        // ---- Output format. ----
        if ((strcmp(argv[arg], "-o") == 0) || (strcmp(argv[arg], "--out-format") == 0)) {
            if (!confirm_remaining_arguments(argv[arg], 1, arg, argc)) {
                goto exit;
            }
            arg += 1;
            if (!parse_out_format(argv[arg], &out_format)) {
                goto exit;
            }
            continue;
        }
        if (strstr(argv[arg], "--out-format=") == argv[arg]) {
            if (!parse_out_format(strchr(argv[arg], '=')+1, &out_format)) {
                goto exit;
            }
            continue;
        }

        // ---- Border width. ----
        if ((strcmp(argv[arg], "-p") == 0) || (strcmp(argv[arg], "--padding") == 0)) {
            if (!confirm_remaining_arguments(argv[arg], 1, arg, argc)) {
                goto exit;
            }
            arg += 1;
            padding = strtol(argv[arg], NULL, 0);
            continue;
        }
        if (strstr(argv[arg], "--padding=") == argv[arg]) {
            padding = strtol(strchr(argv[arg], '=')+1, NULL, 0);
            continue;
        }

        // ---- Help ----
        if ((strcmp(argv[arg], "--help") == 0) || (strcmp(argv[arg], "-h") == 0)) {
            usage();
            ret = EXIT_SUCCESS;
            goto exit;
        }

        sources[source_count] = argv[arg];
        source_count++;
    }

    if (source_count == 0) {
        printf("Error: No sources specified.\n");
        usage();
        goto exit;
    }

    // Convert each source file.
    for (uint32_t i = 0; i < source_count; ++i) {
        rgb_to_yuv(sources[i], out_format, padding);
    }

    ret = EXIT_SUCCESS;
exit:
    free(sources);
    return ret;
}

