/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is a test framework for NEON YUV-to-ARGB code.
 *
 * The Initial Developer of the Original Code is ARM Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jacob Bramley <Jacob.Bramley@arm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <malloc.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>

#include "runner.h"

#include "framework.h"
#include "bmp_writer.h"

// A null scalar that does nothing, to check timing function overheads.
void __attribute__((noinline,unused)) ScaleYCbCrToRGB565_Bilinear_NULL(
        uint8_t const * y,
        uint8_t const * u,
        uint8_t const * v,
        void          * rgb,
        int             src_left,
        int             src_top,
        int             src_width,
        int             src_height,
        int             dst_width,
        int             dst_height,
        int             y_pitch,    // Source row-stride.
        int             uv_pitch,   // Source row-stride for U and V.
        int             rgb_pitch,  // Destination row-stride.
        format_t        src_format)
{
    // Do nothing.
    return;
}

// A simple C implementation of a YUV-to-RGB scaler, for output reference.
void __attribute__((noinline,unused)) ScaleYCbCrToRGB565_Bilinear_C(
        uint8_t const * y,
        uint8_t const * u,
        uint8_t const * v,
        void          * rgb,
        int             src_left,
        int             src_top,
        int             src_width,
        int             src_height,
        int             dst_width,
        int             dst_height,
        int             y_pitch,    // Source row-stride.
        int             uv_pitch,   // Source row-stride for U and V.
        int             rgb_pitch,  // Destination row-stride.
        format_t        src_format)
{
    float src_dx;
    float src_dy;

    float src_x0;
    float src_y0;
    float src_x;
    float src_y;

    // For down-scaling, ensure that we start sampling with an offset.
    // Otherwise, down-scaling by integral factors will effectively result in
    // nearest filtering.
    // src_dx = src_width/dst_width, src_x0 = (src_dx-1)/2
    //
    // For up-scaling, ensure that we don't sample beyond the edge of the
    // source image.
    // src_dx = (src_width-1)/(dst_width-1), src_x0 = 0

    if (dst_width < src_width) {
        // Down-scale.
        src_dx = (float)(src_width) / (float)(dst_width);
        src_x0 = (src_dx - 1.0f) / 2.0f;
    } else {
        // Up-scale.
        src_dx = (float)(src_width-1) / (float)(dst_width-1);
        src_x0 = 0.0f;
    }

    // Do the same for the vertical direction.

    if (dst_height < src_height) {
        // Down-scale.
        src_dy = (float)(src_height) / (float)(dst_height);
        src_y0 = (src_dy - 1.0f) / 2.0f;
    } else {
        // Up-scale.
        src_dy = (float)(src_height-1) / (float)(dst_height-1);
        src_y0 = 0.0f;
    }

    // Adjust source pointers past the padding section.
    y += src_left + y_pitch * src_top;
    u += (src_left/2) + uv_pitch * (src_top/2);
    v += (src_left/2) + uv_pitch * (src_top/2);

    // Convert the image.
    src_y = src_y0;
    for (uint32_t dst_y = 0; dst_y < dst_height; dst_y++, src_y += src_dy) {
        uint16_t *  rgb_row = (uint16_t*)rgb + dst_y * rgb_pitch;
        src_x = src_x0;
        for (uint32_t dst_x = 0; dst_x < dst_width; dst_x++, src_x += src_dx) {
            // We sample four source pixels for each destination pixel.
            // This is not designed to be an efficient implementation, but
            // rather is designed to make the scaling and conversion procedure
            // clear and easy to see.
            // Source co-ordinates.
            uint32_t    src_x_0 = (uint32_t)floorf(src_x);
            uint32_t    src_x_1 = src_x_0 + 1;
            uint32_t    src_y_0 = (uint32_t)floorf(src_y);
            uint32_t    src_y_1 = src_y_0 + 1;
            // Source samples.
            float       src_00;
            float       src_01;
            float       src_10;
            float       src_11;
            // Source sample weightings.
            float       weight_x_0;
            float       weight_x_1;
            float       weight_y_0;
            float       weight_y_1;
            // Scaled source values.
            float       yf;
            float       uf;
            float       vf;
            // Scaled destination values.
            float       rf;
            float       gf;
            float       bf;
            // RGB-565 destination value.
            uint32_t    dst_rgb;
            // YCbCr->RGB conversion co-efficients.
            float const bias_y = 16.0f;
            float const bias_uv = 128.0f;
            float const coef_y = (255.0f/219.0f);
            float const coef_v_red = (255.0f/112.0f)*0.701f;
            float const coef_v_grn = (225.0f/112.0f)*0.701f*(0.299f/0.587f);
            float const coef_u_grn = (255.0f/112.0f)*0.886f*(0.114f/0.587f);
            float const coef_u_blu = (255.0f/112.0f)*0.886f;
            // Ensure that we don't over-read.
            if (src_x_1 >= src_width) {
                src_x_1 = src_x_0; 
                assert((uint32_t)floorf(src_x) == src_width-1);
                assert((float)(src_width-1) - src_x < 0.01f);
            }
            if (src_y_1 >= src_height) {
                src_y_1 = src_y_0; 
                assert((uint32_t)floorf(src_y) == src_height-1);
                assert((float)(src_height-1) - src_y < 0.01f);
            }
            // Calculate the sample weightings.
            weight_y_1 = src_y - floorf(src_y);
            weight_y_0 = 1.0f - weight_y_1;
            weight_x_1 = src_x - floorf(src_x);
            weight_x_0 = 1.0f - weight_x_1;
            // Scale Y.
            src_00 = weight_y_0 * weight_x_0 * (float)y[src_x_0 + src_y_0 * y_pitch];
            src_01 = weight_y_0 * weight_x_1 * (float)y[src_x_1 + src_y_0 * y_pitch];
            src_10 = weight_y_1 * weight_x_0 * (float)y[src_x_0 + src_y_1 * y_pitch];
            src_11 = weight_y_1 * weight_x_1 * (float)y[src_x_1 + src_y_1 * y_pitch];
            yf = src_00 + src_01 + src_10 + src_11;
            // Use nearest-sampling to obtain U and V.
            uf = (float)u[(src_x_0/2) + (src_y_0/2)*(uv_pitch)];
            vf = (float)v[(src_x_0/2) + (src_y_0/2)*(uv_pitch)];
            // Apply bias values.
            yf -= bias_y;
            uf -= bias_uv;
            vf -= bias_uv;
            // Convert to RGB.
            rf = (coef_y * yf) +                     (coef_v_red * vf);
            gf = (coef_y * yf) - (coef_u_grn * uf) - (coef_v_grn * vf);
            bf = (coef_y * yf) + (coef_u_blu * uf)                    ;
            // Clamp to the range 0-255.
            rf = fminf(255.0f, fmaxf(0.0f, rf));
            gf = fminf(255.0f, fmaxf(0.0f, gf));
            bf = fminf(255.0f, fmaxf(0.0f, bf));
            // Store the result in the destination image.
            dst_rgb  = (((uint32_t)rf) << 8) & 0xf800;
            dst_rgb |= (((uint32_t)gf) << 3) & 0x07e0;
            dst_rgb |= (((uint32_t)bf) >> 3) & 0x001f;
            rgb_row[dst_x] = dst_rgb;
        }
    }
}

// Note that this function uses malloc to allocate space for y, u and v, and
// this space must be freed.
int load_ycbcr(options_t * options, uint8_t ** y, uint8_t ** u, uint8_t ** v)
{
    int         ret = 0;
    uint32_t    w = options->src_width + options->pad_left + options->pad_right;
    uint32_t    h = options->src_height + options->pad_top + options->pad_bottom;
    FILE *      in = NULL;

    uint8_t *   y_aligned;
    uint8_t *   u_aligned;
    uint8_t *   v_aligned;

    *y = malloc(16 + w * h);
    *u = malloc(16 + (w*h+3)/4);
    *v = malloc(16 + (w*h+3)/4);
    OOM_CHECK(*y);
    OOM_CHECK(*u);
    OOM_CHECK(*v);

    y_aligned = alignUp(*y, 16);
    u_aligned = alignUp(*u, 16);
    v_aligned = alignUp(*v, 16);

    in = fopen(options->src_file, "rb");
    if (!in) {
        printf("Error: Could not open file \"%s\".\n", options->src_file);
        goto exit;
    }

    switch (options->format) {
      case FMT_IYUV:    // Equivalent to FMT_I420.
        // Planar: 8-bit Y, 8-bit 2x2-sampled U, 8-bit 2x2-sampled V.
        fread(y_aligned, 1, w*h, in);
        fread(u_aligned, 1, (w*h+3)/4, in);
        fread(v_aligned, 1, (w*h+3)/4, in);
        // The only difference between IYUV/I420 and YV12 is the order of
        // the U and V planes in storage, so from here onwards we can treat
        // them as the same format.
        options->format = FMT_YV12;
        break;
      case FMT_YV12:
        // Planar: 8-bit Y, 8-bit 2x2-sampled V, 8-bit 2x2-sampled U.
        fread(y_aligned, 1, w*h, in);
        fread(v_aligned, 1, (w*h+3)/4, in);
        fread(u_aligned, 1, (w*h+3)/4, in);
        break;
      default:
        printf("Error: Support for loading that input format has not been implemented.\n");
        goto exit;
    }
    ret = 1;
exit:
    if (in) {
        fclose(in);
    }
    return ret;
}

void unload_ycbcr(uint8_t * y, uint8_t * u, uint8_t * v)
{
    free(y);
    free(u);
    free(v);
}

// Draw a transparency grid.
void draw_background_rgb565(options_t * options, void * canvas)
{
    // Components:            Red                  Green                Blue
    uint16_t const  dark =  ((0x55<<8)&0xf800) | ((0x55<<3)&0x07e0) | ((0x55>>3)&0x001f);
    uint16_t const  light = ((0x77<<8)&0xf800) | ((0x77<<3)&0x07e0) | ((0x77>>3)&0x001f);
    uint16_t const  alert = ((0x00<<8)&0xf800) | ((0xff<<3)&0x07e0) | ((0x00>>3)&0x001f);
    uint32_t        dst_top;
    uint32_t        dst_right;
    uint32_t        dst_bottom;
    uint32_t        dst_left;

    uint16_t * rgb = canvas;

    for (uint32_t y = 0; y < options->dst_height; y++) {
        for (uint32_t x = 0; x < options->dst_width; x++) {
            if ((y & 0x10) ^ (x & 0x10)) {
                rgb[x+y*options->dst_width] = light;
            } else {
                rgb[x+y*options->dst_width] = dark;
            }
        }
    }

    // Draw a rectangle the overlaps the target area by 1px, to check for
    // off-by-one over-writing or under-writing.
    dst_top = options->dst_y;
    dst_right = options->dst_x + options->src_width * options->scale_x;
    dst_bottom = options->dst_y + options->src_height * options->scale_y;
    dst_left = options->dst_x;
    if (dst_top != 0) {
        dst_top -= 1;
    }
    if (dst_right < (options->dst_width-1)) {
        dst_right++;
    }
    if (dst_bottom < (options->dst_height-1)) {
        dst_bottom++;
    }
    if (dst_left != 0) {
        dst_left -= 1;
    }
    for (uint32_t y = dst_top; y < dst_bottom; y++) {
        for (uint32_t x = dst_left; x < dst_right; x++) {
            rgb[x+y*options->dst_width] = alert;
        }
    }
}

// Run the list of tests.
int run(options_t * options)
{
    int         ret = 0;
    uint32_t    runs = options->runs;
    uint32_t    skips = 1;          // Don't count the first run when benchmarking.

    uint8_t *   y = NULL;
    uint8_t *   u = NULL;
    uint8_t *   v = NULL;
    void *      dst_base = NULL;

    uint8_t *   y_aligned;
    uint8_t *   u_aligned;
    uint8_t *   v_aligned;

    void *      dst;
    uint32_t    dst_width;
    uint32_t    dst_height;

    uint32_t    src_full_width =    // Includes padding.
        options->src_width + options->pad_left + options->pad_right;

    // Output file name.
    char *      dst_name = NULL;

    int         seed = time(NULL);
    if (options->verbose) {
        printf("  Random seed initialized to %d.\n", seed);
    }
    srand(seed);

    // Allocate a buffer for the result.
    dst_base = malloc(sizeof(uint16_t) * options->dst_width * options->dst_height);
    OOM_CHECK(dst_base);

    // Read the input file.
    if (!load_ycbcr(options, &y, &u, &v)) {
        goto exit;
    }

    y_aligned = alignUp(y, 16);
    u_aligned = alignUp(u, 16);
    v_aligned = alignUp(v, 16);

    // Adjust the destination start pointer and size to the required positions.
    dst = (char*)dst_base + sizeof(uint16_t) * (options->dst_x + options->dst_width * options->dst_y);
    dst_width = options->src_width * options->scale_x;
    dst_height = options->src_height * options->scale_y;

    dst_name = malloc(options->base_name_len + 128);
    OOM_CHECK(dst_name);

    // Run each test <skips> times without timing, and save the output images.
    for (uint32_t i = 0; i < skips; i++) {
        for (uint32_t test = 0; test < test_count; ++test) {
            draw_background_rgb565(options, dst_base);
            tests[test].func(   
                    y_aligned, u_aligned, v_aligned, dst,
                    options->pad_left,
                    options->pad_top,
                    options->src_width,
                    options->src_height,
                    dst_width,
                    dst_height,
                    src_full_width,
                    src_full_width/2,
                    options->dst_width,
                    options->format);
            sprintf(dst_name,
                    "%s-%s-%fx%f.bmp",
                    options->base_name,
                    tests[test].file_name_tag,
                    options->scale_x,
                    options->scale_y);
            if (!write_bmp(dst_name, dst_base, options->dst_width, options->dst_height, 16)) {
                goto exit;
            }
        }
    }

    // Run each test <runs> times with timing.
    for (uint32_t i = 0; i < runs; i++) {
        uint32_t list[test_count];
        if (options->verbose) {
            printf("Running permutation %u.\n", i);
        }
        // Initialize the list in sequential order.
        for (uint32_t test = 0; test < test_count; ++test) {
            list[test] = test;
        }
        // Shuffle the list.
        for (uint32_t test = 0; test < test_count; ++test) {
            uint32_t test2 = rand() % test_count;
            uint32_t swap;
            swap = list[test2];
            list[test2] = list[test];
            list[test] = swap;
        }
        // Time each test in the list.
        for (uint32_t perm = 0; perm < test_count; ++perm) {
            struct timespec start;
            struct timespec end;
            double          start_s;
            double          end_s;
            uint32_t const  test = list[perm];

            if (options->verbose) {
                printf("  Running test: %s\n", tests[test].results_tag);
            }
            fflush(stdout);
            if (clock_gettime(CLOCK_REALTIME, &start) != 0) {
                printf("Error: Could not get the timer value.");
                goto exit;
            }
            tests[test].func(   
                    y_aligned, u_aligned, v_aligned, dst,
                    options->pad_left,
                    options->pad_top,
                    options->src_width,
                    options->src_height,
                    dst_width,
                    dst_height,
                    src_full_width,
                    src_full_width/2,
                    options->dst_width,
                    options->format);
            if (clock_gettime(CLOCK_REALTIME, &end) != 0) {
                printf("Error: Could not get the timer value.\n");
                goto exit;
            }

            start_s = start.tv_sec + (double)start.tv_nsec / 1000000000.0;
            end_s =     end.tv_sec + (double)  end.tv_nsec / 1000000000.0;
            if(end_s < start_s) {
                // Sometimes, clock_gettime ticks backwards. This is probably a
                // bug, except that it happens on a couple of systems. In any
                // case, if it happens here, ignore it and run the loop again.
                //if (options->verbose) {
                    printf("Warning: clock_gettime went backwards! Time=%f\n",
                            end_s - start_s);
                //}
                --perm;
                continue;
            }
            tests[test].time += end_s - start_s;
        }
    }

    // Report results.
    for (uint32_t test = 0; test < test_count; ++test) {
        double time_ms;
        double time_rel;
        time_ms = 1000.0 * tests[test].time / (double)runs;
        time_rel = 100.0 * tests[test].time / test_base->time;
        // dst_name = <base_name>-<filter>-<implementation>-<scale_x>x<scale_y>.bmp
        sprintf(dst_name,
                "%s-%s-%fx%f.bmp",
                options->base_name,
                tests[test].file_name_tag,
                options->scale_x,
                options->scale_y);
        if (options->runs) {
            printf("  Method: %-32s   Average time: %7.3fms   Relative time: %5.1f%%\n", tests[test].results_tag, time_ms, time_rel);
        }
        if ((options->verbose) || (!runs)) {
            printf("  Output file name: %s\n", dst_name);
        }
    }

    // --------
    ret = 1;
exit:
    free(dst_name);
    free(dst_base);
    unload_ycbcr(y,u,v);
    return ret;
}

